#ifndef __X86_MOO__
#define __X86_MOO__

#include <inc/adb.h>
#include <inc/sig.h>

struct CPUX86State;

typedef enum {
  X86_EXIT_BAD_INSN,
  X86_EXIT_REQUESTED,
  X86_EXIT_THRESHOLD,
  X86_EXIT_TEMPORARY,
} x86exit_e;

#define X86_EFLAGS_BUF_SIZE 25

#define X86_EFLAGS_CF	0x00000001 /* Carry Flag */
#define X86_EFLAGS_2O   0x00000002
#define X86_EFLAGS_PF	0x00000004 /* Parity Flag */
#define X86_EFLAGS_3Z   0x00000008
#define X86_EFLAGS_AF	0x00000010 /* Auxillary carry Flag */
#define X86_EFLAGS_5Z   0x00000020
#define X86_EFLAGS_ZF	0x00000040 /* Zero Flag */
#define X86_EFLAGS_SF	0x00000080 /* Sign Flag */
#define X86_EFLAGS_TF	0x00000100 /* Trap Flag */
#define X86_EFLAGS_IF	0x00000200 /* Interrupt Flag */
#define X86_EFLAGS_DF	0x00000400 /* Direction Flag */
#define X86_EFLAGS_OF	0x00000800 /* Overflow Flag */
#define X86_EFLAGS_IOPL	0x00003000 /* IOPL mask */
#define X86_EFLAGS_NT	0x00004000 /* Nested Task */
#define X86_EFLAGS_15Z  0x00008000
#define X86_EFLAGS_RF	0x00010000 /* Resume Flag */
#define X86_EFLAGS_VM	0x00020000 /* Virtual Mode */
#define X86_EFLAGS_AC	0x00040000 /* Alignment Check */
#define X86_EFLAGS_VIF	0x00080000 /* Virtual Interrupt Flag */
#define X86_EFLAGS_VIP	0x00100000 /* Virtual Interrupt Pending */
#define X86_EFLAGS_ID	0x00200000 /* CPUID detection flag */

enum {
  X86_PREFIX_ES = 1,
  X86_PREFIX_CS = 2,
  X86_PREFIX_SS = 4,
  X86_PREFIX_DS = 8,
  X86_PREFIX_FS = 16,
  X86_PREFIX_GS = 32,
  X86_PREFIX_DATA = 64,
  X86_PREFIX_ADDR = 128,
  X86_PREFIX_LOCK = 256,
  X86_PREFIX_REPNE = 512,
  X86_PREFIX_REP = 1024,
};

extern void         X86_init (void);
extern void         X86_pre_init (void);
extern void         X86_catch_VGA (int);
extern x86exit_e    X86_run (struct CPUX86State *, int threshold);
extern x86exit_e    X86_run32 (regs86 *r, int threshold);
extern x86exit_e    X86_confirm (struct CPUX86State *, int threshold);
extern x86exit_e    X86_confirm32 (regs86 *r, int threshold);
extern void         X86_dump_current (void);
extern char       * X86_efltoa (ui eflags, char m[X86_EFLAGS_BUF_SIZE]);
extern void         X86_stop (void);
extern const char * X86_reg_name (ub nreg);
extern const char * X86_seg_name (ub nseg);

#endif /* __X86_MOO__ */
