open Zz
open Types
open Format

let indent = 4
let confirm = ref false

(* needs more work *)
let decl_init = ref false
let read_const = ref false

let cstring_of_size = function
    S8 -> "BYTE"
  | S16 -> "WORD"
  | S32 -> "DWORD"

let string_of_cast size = function
    Cast(Iunsigned, _) -> "(" ^ cstring_of_size size ^ ") "
  | Cast(Isigned, _) -> "(S" ^ cstring_of_size size ^ ") "
  | otherwise -> ""

let pp_size ppf size = fprintf ppf "%s" (string_of_size size)
let pp_sign ppf sign = fprintf ppf "%s" (string_of_sign sign)
let pp_cast ppf = function
    Cast (sign, size) ->
      fprintf ppf "cast(%c%c)"
	(char_of_sign sign) (char_of_size size)
  | Raw -> fprintf ppf "nocast"

let pp_if_else cond ppf pp_if_else =
  pp_open_vbox ppf indent;
  pp_print_string ppf "if (";
  pp_print_string ppf cond;
  pp_print_string ppf ") {";
  pp_print_cut ppf ();
  pp_if_else ppf ~if_clause:true;
  pp_close_box ppf ();
  pp_print_cut ppf ();
  pp_print_string ppf "}";
  pp_print_cut ppf ();
  pp_open_vbox ppf indent;
  pp_print_string ppf "else {";
  pp_print_cut ppf ();
  pp_if_else ppf ~if_clause:false;
  pp_close_box ppf ();
  pp_print_cut ppf ();
  pp_print_string ppf "}"

let pp_decl ppf ~name ~size ~cast ~is_read =
  if !decl_init then
    let ctype = ctype_name (ctype_code size cast) in
    if !read_const && is_read
    then fprintf ppf "const %s %s" ctype name
    else fprintf ppf "%s %s" ctype name
  else
    fprintf ppf "%s" name

let pp
    ~rm_state
    ~is_read
    ~rm_is_reg
    ~assume
    ppf
    ((pos, zz, size, cast), vname) =

  let inv = assume = Anone or assume = Afree in

  let pp_decl ppf name =
    if !decl_init then
      let ctype = ctype_name (ctype_code size cast) in
      if !read_const && is_read
      then fprintf ppf "const %s %s" ctype name
      else fprintf ppf "%s %s" ctype name
    else
      fprintf ppf "%s" name
  in

  let pp_decl_hard hard_size ppf name =
    if !decl_init then
      let ctype = cstring_of_size hard_size in
      if !read_const && is_read
      then fprintf ppf "const %s %s" ctype name
      else fprintf ppf "%s %s" ctype name
    else
      fprintf ppf "%s" name
  in

  let n = succ pos in
  let v = if size = S32 then "DWORD" else "WORD" in
  let cast = string_of_cast size cast in

  let is_rm_local () =
    if is_read then
      match rm_state with
      | Rm_reg -> true
      | Rm_r -> true
      | Rm_w -> failwith "read rm requested with modrm write only"
      | Rm_wf | Rm_rw -> false
      | Rm_nil -> failwith "read rm requested but there is no spoon"
    else
      match rm_state with
      | Rm_reg -> true
      | Rm_r -> failwith "read rm requested with modrm write only"
      | Rm_w -> true
      | Rm_wf | Rm_rw -> false
      | Rm_nil -> failwith "read rm requested but there is no spoon"
  in

  let init_rm ppf is_mem =
    let is_local = is_rm_local () in
    let do_decl () =
      if is_local then
	if !decl_init
	then
	  if !read_const
	  then fprintf ppf "const DWORD "
	  else fprintf ppf "DWORD "
	else
	  fprintf ppf "DWORD rm;@,";
    in
    if is_read then begin
      do_decl ();
      if is_mem then
	fprintf ppf "rm = LEA (env, modrm);"
      else
	fprintf ppf "rm = GET_MODRM_RM (modrm);"
    end else begin
      if is_local then begin
	do_decl ();
	if is_mem then
	  fprintf ppf "rm = LEA (env, modrm);"
	else
	  fprintf ppf "rm = GET_MODRM_RM (modrm);"
      end else fprintf ppf "/* NOT LOCAL %d %d */"
	  (Obj.magic rm_state)
	  (Obj.magic Rm_r);
    end
  in

  let imm_maybe_init_rm ~is_Ox =
    if not is_read && not is_Ox
    then failwith "write to immediate"
    else begin
      begin match rm_state with
      | Rm_w -> failwith "modrm state inconsistent w.r.t immediate operands"
      | Rm_wf ->
	  begin match assume with
	    Anone ->
	      let pp_if ppf ~if_clause = init_rm ppf if_clause in
	      pp_if_else "IS_MODRM_MEM (modrm)" ppf pp_if
	  | Amem ->
	      init_rm ppf true
	  | Areg ->
	      init_rm ppf false
	  | Afree -> ()
	  end
      | otherwise -> ()
      end;
      if not is_Ox then fprintf ppf "@,";
    end
  in

  let const_reg typ name =
    if inv then begin
      if is_read then
	fprintf ppf "%a = %sREAD_%s_REG (env, _REG_%s);"
	  pp_decl vname cast typ name
      else
	fprintf ppf "WRITE_%s_REG (env, _REG_%s, %s);" typ name vname
    end
  in

  let const_seg name =
    if inv then begin
      if is_read then
	fprintf ppf "%a = READ_SEGMENT_REG (env, _SEG_%s);"
	  pp_decl vname name
      else
	fprintf ppf "WRITE_SEGMENT_REG (env, _SEG_%s, %s);" name vname
    end
  in

  let modrm_reg typ =
    if inv then begin
      let r =
	if rm_is_reg
	then "GET_MODRM_RM (modrm)"
	else "GET_MODRM_REG (modrm)"
      in
      if is_read then
	fprintf ppf "%a = %sREAD_%s_REG (env, %s);"
	  pp_decl vname cast typ r
      else
	fprintf ppf "WRITE_%s_REG (env, %s, %s);" typ r vname
    end
  in

  let _Ix typ =
    let cond =
      if rm_state = Rm_wf or rm_state = Rm_rw
      then assume != Afree
      else assume = Anone
    in
    if cond then begin
      imm_maybe_init_rm ~is_Ox:false;
      fprintf ppf "%a = %sFETCH_%s (env);" pp_decl vname cast typ
    end
  in

  let _Jx typ =
    imm_maybe_init_rm ~is_Ox:false;
    fprintf ppf "%a = %sFETCH_%s (env);" pp_decl vname cast typ in

  let _Ox typ =
    let pp_if ppf ~if_clause =
      if if_clause then begin
	fprintf ppf "PTR addr;@,@,";
	fprintf ppf "addr = FETCH_DWORD (env);@,";
	if is_read then
	  fprintf ppf "%a = READ_%s_MEM (env, addr);" pp_decl vname typ
	else
	  fprintf ppf "WRITE_%s_MEM (env, addr, %s);" typ vname
      end else begin
	fprintf ppf "PTR addr;@,@,";
	fprintf ppf "addr = FETCH_WORD (env);@,";
	if is_read then
	  fprintf ppf "%a = READ_%s_MEM (env, addr);" pp_decl vname typ
	else
	  fprintf ppf "WRITE_%s_MEM (env, addr, %s);" typ vname
      end
    in
    imm_maybe_init_rm ~is_Ox:true;
    pp_if_else "IS_ADDRESS_SIZE_32 (env)" ppf pp_if
  in

  let _Ex typ =
    let pp_if ppf ~if_clause =
      fprintf ppf "%a@," init_rm if_clause;
      if if_clause then begin
	if is_read then
	  fprintf ppf "%a = %sREAD_%s_MEM (env, rm);" pp_decl vname cast typ
	else begin
	  if !confirm then
	    fprintf ppf "if (0 == (env->iflags & IF_CONFIRM))@,";
	  fprintf ppf "WRITE_%s_MEM (env, rm, %s);" typ vname
	end
      end else begin
	if is_read then
	  fprintf ppf "%a = %sREAD_%s_REG (env, rm);" pp_decl vname cast typ
	else
	  fprintf ppf "WRITE_%s_REG (env, rm, %s);" typ vname
      end
    in
    match assume with
      Anone ->
	pp_if_else "IS_MODRM_MEM (modrm)" ppf pp_if
    | Amem ->
	pp_if ppf true
    | Areg ->
	pp_if ppf false
    | Afree -> ()
  in

  let _Rx typ =
    let pp_if ppf ~if_clause =
      if if_clause then begin
	pp_print_string ppf "EXCEPTION (env, 6, 0); return;";
      end else begin
	fprintf ppf "%a" init_rm false;
	if is_read then
	  fprintf ppf "%a = %sREAD_%s_REG (env, rm);" pp_decl vname cast typ
	else
	  fprintf ppf "WRITE_%s_REG (env, rm, %s);" typ vname
      end
    in
    match assume with
      Anone ->
	pp_if_else "IS_MODRM_MEM (modrm)" ppf pp_if
    | Amem ->
	pp_if ppf true
    | Areg ->
	pp_if ppf false
    | Afree -> ()
  in

  let _Mx typ =
    let pp_if ppf ~if_clause =
      if if_clause then begin
	fprintf ppf "%a@," init_rm true;
	fprintf ppf "%a = %sREAD_%s_MEM (env, rm);" pp_decl vname cast typ;
      end else begin
	pp_print_string ppf "EXCEPTION (env, 6, 0); return;";
      end
    in
    match assume with
      Anone ->
	pp_if_else "IS_MODRM_MEM (modrm)" ppf pp_if
    | Amem ->
	pp_if ppf true
    | Areg ->
	pp_if ppf false
    | Afree -> ()
  in

  match zz with
  | AL -> const_reg "BYTE" (Reg.name8 Reg._AL)
  | CL -> const_reg "BYTE" (Reg.name8 Reg._CL)
  | DX -> const_reg "WORD" (Reg.name16 Reg._DX)
  | EAX ->
      const_reg v
	((if size = S32 then Reg.name32 else Reg.name16) Reg._EAX)

  | ES -> const_seg (Reg.nameseg Reg._ES)
  | CS -> const_seg (Reg.nameseg Reg._CS)
  | DS -> const_seg (Reg.nameseg Reg._DS)
  | SS -> const_seg (Reg.nameseg Reg._SS)
  | FS -> const_seg (Reg.nameseg Reg._FS)
  | GS -> const_seg (Reg.nameseg Reg._GS)

  | Regb ->
      if inv then begin
	if is_read then
	  fprintf ppf "%a = %sREAD_BYTE_REG (env, opcode & 7);"
	    pp_decl vname cast
	else
	  fprintf ppf "WRITE_BYTE_REG (env, opcode & 7, %s);" vname
      end

  | Regv ->
      if inv then begin
	if is_read then
	  fprintf ppf "%a = %sREAD_%s_REG (env, opcode & 7);"
	    pp_decl vname cast v
	else
	  fprintf ppf "WRITE_%s_REG (env, opcode & 7, %s);" v vname
      end

  | Ib -> _Ix "BYTE"
  | Iw -> _Ix "WORD"
  | Iv -> _Ix v

  | Jb -> _Jx "BYTE"
  | Jv -> _Jx v

  | Const1 -> if inv then fprintf ppf "%a = 1;" pp_decl vname

  | Rd -> _Rx "DWORD"
  | Rw -> _Rx "WORD"

  | Eb -> _Ex "BYTE"
  | Ew -> _Ex "WORD"
  | Ev -> _Ex v

  | Gv -> modrm_reg v
  | Gb -> modrm_reg "BYTE"
  | Td -> modrm_reg "TEST"
  | Cd -> modrm_reg "CONTROL"
  | Dd -> modrm_reg "DEBUG"
  | Sw -> modrm_reg "SEGMENT"

  | M ->
      let pp_if ppf ~if_clause =
	if if_clause then begin
	  fprintf ppf "%a@," init_rm true;
	  fprintf ppf "%a = rm;@," pp_decl vname;
	end else begin
	  pp_print_string ppf "EXCEPTION (env, 6, 0); return;";
	end
      in
      pp_if_else "IS_MODRM_MEM (modrm)" ppf pp_if

  | Ma ->
      let pp_if ppf ~if_clause =
	if if_clause then begin
	  fprintf ppf "%a@," init_rm true;
	  fprintf ppf "%a_lo = READ_%s_MEM (env, rm);@,"
	    pp_decl vname v;
	  fprintf ppf "%a_hi = READ_%s_MEM (env, rm + %d);"
	    pp_decl vname
	    v (if size = S32 then 4 else 2);
	end else begin
	  pp_print_string ppf "EXCEPTION (env, 6, 0); return;";
	end
      in
      begin match assume with
	Anone ->
	  pp_if_else "IS_MODRM_MEM (modrm)" ppf pp_if
      | Amem ->
	  pp_if ppf true
      | Areg ->
	  pp_if ppf false
      | Afree -> ()
      end

  | Ep | Mp ->
      let pp_if ppf ~if_clause =
	if if_clause then begin
	  fprintf ppf "%a@," init_rm true;
	  fprintf ppf "%a_off = READ_%s_MEM (env, rm);@,"
	    pp_decl vname v;

	  (* XXX *)
	  fprintf ppf "%a_seg = READ_WORD_MEM (env, rm + %d);"
	    (pp_decl_hard S16) vname (if size = S32 then 4 else 2);
	end else begin
	  pp_print_string ppf "EXCEPTION (env, 6, 0); return;";
	end
      in
      begin match assume with
	Anone ->
	  pp_if_else "IS_MODRM_MEM (modrm)" ppf pp_if
      | Amem ->
	  pp_if ppf true
      | Areg ->
	  pp_if ppf false
      | Afree -> ()
      end

  | Ob -> _Ox "BYTE"
  | Ov -> _Ox v

  | Fv ->
      if is_read then
	fprintf ppf "%a = READ_%s_FLAGS (env);" pp_decl vname v
      else
	fprintf ppf "WRITE_%s_FLAGS (env, %s);" v vname

  | Ap ->
      if size = S32 then begin
	fprintf ppf "%a_off = FETCH_DWORD (env);@," pp_decl vname;
	fprintf ppf "%a_seg = FETCH_WORD (env);" (pp_decl_hard S16) vname
      end else begin
	fprintf ppf "%a_off = FETCH_WORD (env);@," pp_decl vname;
	fprintf ppf "%a_seg = FETCH_WORD (env);" (pp_decl_hard S16) vname
      end

  | Xb
  | Xv
  | Yb
  | Yv -> assert false
  | Ms -> failwith "Ms not supported yet"

let fprintf_debug ppf (dep, opcode, mnemonic, zz_list, rvars, wvars) =
  let pp_fmt ppf is_read =
    let pp ppf (var, name) =
      match var with
	Operand(pos, zz, size, cast) ->
	  begin match zz with
	    Ap|Ep|Mp ->
	      fprintf ppf "seg:off=%%x:%%x"
	  | Ma ->
	      fprintf ppf "lo:hi=%%x,%%x"
	  | Ms|Xb|Yb|Xv|Yv -> ()
	  | AL ->
	      fprintf ppf "AL=%%x"
	  | CL ->
	      fprintf ppf "CL=%%x"
	  | DX ->
	      fprintf ppf "DX=%%x"
	  | EAX ->
	      if size = S32 then fprintf ppf "E";
	      fprintf ppf "AX=%%x"
	  | ES|CS|DS|SS|FS|GS ->
	      fprintf ppf "seg=%%x"
	  | Regb ->
	      fprintf ppf "%s=%%x" (Reg.name8 (opcode land 7))
	  | Regv ->
	      if size = S32 then
		fprintf ppf "%s=%%x" (Reg.name32 (opcode land 7))
	      else
		fprintf ppf "%s=%%x" (Reg.name16 (opcode land 7))
	  | Ib|Iw|Iv ->
	      fprintf ppf "imm=%%x"
	  | Jb|Jv ->
	      fprintf ppf "rel=%%x"
	  | Ob|Ov ->
	      fprintf ppf "[O]=%%x"
	  | Const1 ->
	      ()
	  | Sw ->
	      fprintf ppf "sw[%%d]=%%x"
	  | Rw ->
	      fprintf ppf "rw[%%d]=%%x"
	  | Rd ->
	      fprintf ppf "rd[%%d]=%%x"
	  | Eb|Ew|Ev ->
	      fprintf ppf "r/m=%%x"
	  | Gv|Gb ->
	      fprintf ppf "gx[%%d]=%%x"
	  | Td ->
	      fprintf ppf "tr%%d=%%x"
	  | Cd ->
	      fprintf ppf "cr%%d=%%x"
	  | Dd ->
	      fprintf ppf "dr%%d=%%x"
	  | M ->
	      fprintf ppf "M=%%x"
	  | Fv ->
	      fprintf ppf "f=%%x"
	  end
      | Reg (reg, cast) ->
	  fprintf ppf "%s=%%x" (Reg.name reg)
      | Flag ->
	  fprintf ppf "f=%%x"
      | EIP ->
	  fprintf ppf "eip=%%x"
    in
    let rec pp_list = function
	var :: [] ->
	  fprintf ppf "%a" pp var
      | var :: vars ->
	  fprintf ppf "%a " pp var; pp_list vars
      | [] -> ()
    in
    let list = if is_read then rvars else wvars in
    if list != [] then begin
      fprintf ppf " %s: " (if is_read then "in" else "out");
      pp_list list
    end
  in

  let pp_arglist ppf is_read =
    let p = if is_read then "r" else "w" in
    let pp ppf (var, vname) =
      match var with
	Operand(pos, zz, size, cast) ->
	  let pos = pos + 1 in
	  begin match zz with
	    Ap|Ep|Mp ->
	      fprintf ppf "%s_seg, %s_off" vname vname
	  | Ma ->
	      fprintf ppf "%s_hi, %s_lo" vname vname
	  | AL
	  | CL
	  | DX
	  | EAX
	  | ES|CS|DS|SS|FS|GS
	  | Regb
	  | Regv
	  | Ib|Iw|Iv
	  | Jb|Jv
	  | Ob|Ov
	  | Eb|Ew|Ev
	  | M ->
	      fprintf ppf "%s" vname
	  | Fv ->
	      fprintf ppf "%s" vname
	  | Rw
	  | Rd
	  | Sw
	  | Gv
	  | Gb ->
	      fprintf ppf "GET_MODRM_REG (modrm), %s" vname
	  | Td
	  | Cd
	  | Dd ->
	      fprintf ppf "GET_MODRM_REG (modrm), %s" vname
	  | Ms|Xb|Yb|Xv|Yv -> ()
	  | Const1 -> ()
	  end
      | Reg (reg, cast) ->
	  fprintf ppf "%s" vname
      | Flag ->
	  fprintf ppf "%s" vname
      | EIP ->
	  assert false
    in
    let rec pp_list = function
	var :: [] ->
	  fprintf ppf "%a" pp var
      | var :: vars ->
	  fprintf ppf "%a,@ " pp var;
	  pp_list vars
      | [] -> ()
    in
    let filt (var, vname) =
      match var with
	Operand (_, (Ms|Xb|Yb|Xv|Yv|Const1), _, _) -> false
      | otherwise -> true
    in
    let list = if is_read then rvars else wvars in
    let list = List.filter filt list in
    if list != [] then begin
      fprintf ppf ",@,";
      pp_list list
    end
  in

(*      fprintf ppf*)
(*	"fprintf (stderr, \"%%08x %%3x %%s %%s\\n\", (DWORD) env->segs[_SEG_CS].base + env->eip, %d, \"%s\", \"%s\");@,"*)
(*	opcode mnemonic (String.concat "," zz_names);*)
  let zz_names =
    List.map begin fun zz ->
      sprintf "%2s"  (Zz.spec_name dep zz)
    end zz_list
  in

  fprintf ppf
    "fprintf @[(turtle_log, \"%%08x %3x %6s %8s:%a%a\\n\",@,(DWORD) env->segs[_SEG_CS].base + env->eip%a%a);@]"
    opcode mnemonic (String.concat "," zz_names)
    pp_fmt true pp_fmt false pp_arglist true pp_arglist false;
