#ifndef __INT_MOO__
#define __INT_MOO__

#include <inc/adb.h>
#include <inc/sig.h>

enum {
  INT_REG_GS = 0,
  INT_REG_FS,
  INT_REG_ES,
  INT_REG_DS,
  INT_REG_EDI,
  INT_REG_ESI,
  INT_REG_EBP,
  INT_REG_ESP,
  INT_REG_EBX,
  INT_REG_EDX,
  INT_REG_ECX,
  INT_REG_EAX,
  INT_REG_EIP,
  INT_REG_CS,
  INT_REG_EFL,
  INT_NREGS
};

typedef ui intregs[INT_NREGS];
#define INT_ARGS intregs r

typedef void (*int_handler)(INT_ARGS);

#define EAX(r) (r)[INT_REG_EAX]
#define EBX(r) (r)[INT_REG_EBX]
#define ECX(r) (r)[INT_REG_ECX]
#define EDX(r) (r)[INT_REG_EDX]
#define ESI(r) (r)[INT_REG_ESI]
#define EDI(r) (r)[INT_REG_EDI]
#define EBP(r) (r)[INT_REG_EBP]
#define ESP(r) (r)[INT_REG_ESP]
#define EIP(r) (r)[INT_REG_EIP]
#define EFL(r) (r)[INT_REG_EFL]
#define CS(r) (r)[INT_REG_CS]
#define DS(r) (r)[INT_REG_DS]
#define ES(r) (r)[INT_REG_ES]
#define FS(r) (r)[INT_REG_FS]
#define GS(r) (r)[INT_REG_GS]
#define SS(r) (r)[INT_REG_SS]

#define AL(r) (EAX(r) & 0xff)
#define BL(r) (EBX(r) & 0xff)
#define CL(r) (ECX(r) & 0xff)
#define DL(r) (EDX(r) & 0xff)

#define AH(r) ((EAX(r) >> 8) & 0xff)
#define BH(r) ((EBX(r) >> 8) & 0xff)
#define CH(r) ((ECX(r) >> 8) & 0xff)
#define DH(r) ((EDX(r) >> 8) & 0xff)

#define AX(r) (EAX(r) & 0xffff)
#define BX(r) (EBX(r) & 0xffff)
#define CX(r) (ECX(r) & 0xffff)
#define DX(r) (EDX(r) & 0xffff)
#define SI(r) (ESI(r) & 0xffff)
#define DI(r) (EDI(r) & 0xffff)
#define BP(r) (EBP(r) & 0xffff)
#define SP(r) (ESP(r) & 0xffff)

#define SET_AL(r, val) EAX(r) = (EAX(r) & ~0xff) | ((ub) (val))
#define SET_BL(r, val) EBX(r) = (EBX(r) & ~0xff) | ((ub) (val))
#define SET_CL(r, val) ECX(r) = (ECX(r) & ~0xff) | ((ub) (val))
#define SET_DL(r, val) EDX(r) = (EDX(r) & ~0xff) | ((ub) (val))

#define SET_AH(r, val) EAX(r) = (EAX(r) & ~0xff00) | (((ub) (val)) << 8)
#define SET_BH(r, val) EBX(r) = (EBX(r) & ~0xff00) | (((ub) (val)) << 8)
#define SET_CH(r, val) ECX(r) = (ECX(r) & ~0xff00) | (((ub) (val)) << 8)
#define SET_DH(r, val) EDX(r) = (EDX(r) & ~0xff00) | (((ub) (val)) << 8)

#define SET_AX(r, val) EAX(r) = (EAX(r) & ~0xffff) | ((us) (val))
#define SET_BX(r, val) EBX(r) = (EBX(r) & ~0xffff) | ((us) (val))
#define SET_CX(r, val) ECX(r) = (ECX(r) & ~0xffff) | ((us) (val))
#define SET_DX(r, val) EDX(r) = (EDX(r) & ~0xffff) | ((us) (val))
#define SET_SI(r, val) ESI(r) = (ESI(r) & ~0xffff) | ((us) (val))
#define SET_DI(r, val) EDI(r) = (EDI(r) & ~0xffff) | ((us) (val))
#define SET_BP(r, val) EBP(r) = (EBP(r) & ~0xffff) | ((us) (val))
#define SET_SP(r, val) ESP(r) = (ESP(r) & ~0xffff) | ((us) (val))

#define SET_ZF(r) EFL (r) |= 64;
#define CLR_ZF(r) EFL (r) &= ~64;

#define SET_CF(r) EFL(r) |= 1;
#define CLR_CF(r) EFL(r) &= ~1;

extern void INT_init           (void);
extern void INT_set            (int nint, int_handler handler);
extern void INT_get_rm         (int nint, us *seg, us *off);
extern void INT_set_rm         (int nint, us seg, us off);
extern void INT_get_pm         (int nint, us *sel, ui *off);
extern void INT_set_pm         (int nint, us sel, ui off);
extern void INT_enable         (void);
extern void INT_disable        (void);
extern int  INT_is_enabled     (void);
extern void INT_simulate       (int nint, regs86 *r);
extern void INT_simulate_real  (int nint, intregs r);
extern void INT_log_regs       (INT_ARGS);

extern void I10 (INT_ARGS);
extern void I11 (INT_ARGS);
extern void I12 (INT_ARGS);
extern void I15 (INT_ARGS);
extern void I16 (INT_ARGS);
extern void I1A (INT_ARGS);
extern void I20 (INT_ARGS);
extern void I21 (INT_ARGS);
extern void I2F (INT_ARGS);
extern void I29 (INT_ARGS);
extern void I31 (INT_ARGS);
extern void I33 (INT_ARGS);
extern void I4B (INT_ARGS);

#define UNHANDLED_INT(args) do {                \
  log ("unhandled situation in %s %s:%d\n",     \
       __func__, __FILE__, __LINE__);           \
  INT_log_regs (args);                          \
  SIG_break (1);                                \
} while (0)

#define UNIMPLEMENTED_INT(args) do {            \
  log ("unimplemented in %s %s:%d\n",           \
       __func__, __FILE__, __LINE__);           \
  INT_log_regs (args);                          \
} while (0)

#endif /* __INT_MOO__ */
