#ifndef __LOG_MOO__
#define __LOG_MOO__

#include <inc/adb.h>

enum LOG_LEVEL {
  LOG_NONE,
  LOG_WARN,
  LOG_INFO,
  LOG_VERBOSE,
  LOG_EXTRA_VERBOSE,
  LOG_DEBUG
};

extern void LOG_pre_init (void);
extern void LOG_init (void);

extern void LOG_log (const char *fmt, ...)
#if defined __GNUC__ || defined __ICC
 __attribute__ ((format (printf, 1, 2)))
#else
#endif
  ;

extern void LOG_log_pref (const char *func,
                          const char *pref,
                          const char *fmt, ...)
#if defined __GNUC__ || defined __ICC
 __attribute__ ((format (printf, 3, 4)))
#else
#endif
  ;

extern void LOG_fail (const char *func, const char *file, int line,
                      const char *fmt, ...)
#if defined __GNUC__ || defined __ICC
 __attribute__ ((format (printf, 4, 5)))
 __attribute__ ((noreturn))
#else
#endif
  ;

extern void LOG_err_fail (const char *func, const char *file, int line,
			  const char *fmt, ...)
#if defined __GNUC__ || defined __ICC
 __attribute__ ((format (printf, 4, 5)))
 __attribute__ ((noreturn))
#else
#endif
  ;

#define _log LOG_log
#define log(...) LOG_log_pref (LOG_PREFIX, __func__, __VA_ARGS__)
#define logn(level, ...) do {                           \
  if ((log_level != LOG_NONE) && (level <= log_level))  \
    LOG_log_pref (LOG_PREFIX, __func__, __VA_ARGS__);   \
} while (0)

#define logn_(level, ...) do {                          \
  if ((log_level != LOG_NONE) && (level <= log_level))  \
    LOG_log (LOG_PREFIX, __VA_ARGS__);                  \
} while (0)

#ifndef lwarn
#define lwarn(...) logn (LOG_WARN, __VA_ARGS__)
#endif

#ifndef linfo
#define linfo(...) logn (LOG_INFO, __VA_ARGS__)
#endif

#ifndef lverbose
#define lverbose(...) logn (LOG_VERBOSE, __VA_ARGS__)
#endif

#ifndef leverbose
#define leverbose(...) logn (LOG_EXTRA_VERBOSE, __VA_ARGS__)
#endif

#ifndef ldebug
#define ldebug(...) logn (LOG_DEBUG, __VA_ARGS__)
#endif

#define Fail_if(cond) do { if (cond) Fail (#cond); } while (0)

#define Fail(...) do {                                  \
  LOG_fail (__func__, __FILE__, __LINE__, __VA_ARGS__); \
} while (0)

#define ERRFail(...) do {                                       \
  LOG_err_fail (__func__, __FILE__, __LINE__, __VA_ARGS__);     \
} while (0)

#define Aldri() Fail ("Me Fail English?\nThats Unpossible!")

#endif /* __LOG_MOO__ */
