#ifndef __SIG_MOO__
#define __SIG_MOO__

#include <inc/adb.h>
#include <inc/reg.h>

struct SIG_mouse_info {
  int off, seg, mask, state, xpos, ypos, hor_mickeys, ver_mickeys;
  int pending;
};

extern void    SIG_pre_init (void);
extern void    SIG_init (void);
extern void    SIG_wait (void);
extern void    SIG_break (int fatal);
extern int     SIG_block (void);
extern void    SIG_unblock (int);
extern void    SIG_start (regs86 *r);
extern void    SIG_log_regs (regs86 *r);
extern void    SIG_vm86 (struct vm86_regs *r, int no_stack);
extern int     SIG_is_in_pic_loop (void);
extern void    SIG_mouse_callback (struct SIG_mouse_info *info);

extern void *SIG_malloc (unsigned int size,
                         char *file, const char *func, int line);

extern void SIG_free (void *ptr,
                      char *file, const char *func, int line);

extern char *SIG_strdup (const char *s, const char *file,
                         const char *func, int line);

extern void *SIG_realloc (void *ptr, unsigned int size,
                          char *file, const char *func, int line);

#define xmalloc(n) SIG_malloc(n,__FILE__,__func__,__LINE__)
#define xfree(ptr) SIG_free(ptr,__FILE__,__func__,__LINE__)
#define xstrdup(s) SIG_strdup(s,__FILE__,__func__,__LINE__)
#define xrealloc(ptr,n) SIG_realloc(ptr,n,__FILE__,__func__,__LINE__)
#define maybe_free(ptr) do{if(ptr){xfree(ptr);ptr=0;}}while(0)

#define UNHANDLED_SIG(r) do {                   \
  log ("unhandled situation in %s %s:%d\n",     \
       __func__, __FILE__, __LINE__);           \
  SIG_log_regs (r);                             \
  SIG_break (1);                                \
} while (0)

#endif /* __SIG_MOO__ */
