#define __ASM__
#include <sys/syscall.h>
#include "inc/red.h"
#include "inc/lbl.h"

#define GLOB(n) .globl G(n); G(n)

        .intel_syntax noprefix

        .code16
        .data

GLOB(int10_redirector):
        cmp     ah, 0x4f
        je      L(handle_vbe_mode_set)

        test    ah, ah
        jnz     L(chain_to_bios)

        pushf
        .byte   0x9a
        .word   RED_INTERRUPT_00 + 0x10
        .word   RED_SEGMENT

L(chain_to_bios):
        .byte   0xea
GLOB(bios_int10_handler):
        .long   0

L(handle_vbe_mode_set):
        .byte   0xea
        .word   RED_INTERRUPT_00 + 0x10
        .word   RED_SEGMENT
GLOB(int10_redirector_end):

        .code32
        .text

GLOB(is_segment_big):
        lar     eax, [esp + 4]
        jz      0f
        mov     eax, -1
        ret
0:
        shr     eax, 22
        and     eax, 1
        ret

GLOB(vm86):
        push    ebx
        mov     eax, SYS_vm86
        mov     ebx, [esp + 8]
        mov     ecx, [esp + 12]
        int     0x80
GLOB(vm86_ret):
        mov     fs, G(fs)
        mov     gs, G(gs)
        pop     ebx
        ret

GLOB(restore_fs_gs):
        mov     fs, G(fs)
        mov     gs, G(gs)
        ret

GLOB(reload_stack):
        movzx   esp, sp
        lret

GLOB(init):
        mov     G(es), es
        mov     G(cs), cs
        mov     G(ds), ds
        mov     G(ss), ss
        mov     G(fs), fs
        mov     G(gs), gs
        lea     eax, [esp + 4]
        mov     G(stack_top), eax
        ret

GLOB(simulate_irq):
        push    ebx
        mov     eax, fs
        mov     ebx, gs
        xchg    eax, [esp + 24]
        xchg    ebx, [esp + 28]
        mov     fs, eax
        mov     gs, ebx

        mov     eax, [esp + 20]
        or      eax, eax
        jz      L(same_stack)

        mov     eax, [esp + 8]
        mov     ebx, [esp + 12]
        mov     ecx, ss
        mov     edx, esp

        lss     esp, [esp + 16]
        push    ecx
        push    edx

        pushf
        push    cs
        push    offset 0f

        push    ebx
        push    eax
        lret
0:
        cld
        pop     edx
        pop     ecx
        mov     ss, ecx
        mov     esp, edx

        mov     fs, [esp + 24]
        mov     gs, [esp + 28]
        pop     ebx
        ret

L(same_stack):
        pushf
        lcall   [esp + 12]
        mov     fs, [esp + 24]
        mov     gs, [esp + 28]
        pop     ebx
        ret

#define ARG(n) ((n)*4)

GLOB(invoke_callback):
        # rm_stack_off 0
        # rm_stack_sel 1
        # data_off     2
        # data_sel     3
        # stack_off    4
        # stack_sel    5
        # proc_off     6
        # proc_sel     7
        pusha

        mov     eax, esp
        lss     esp, [esp + ARG (9+4)]

        push    ds
        push    eax

        mov     edi, [eax + ARG (9+3)]
        mov     es, [edi]
        mov     edi, [eax + ARG (9+2)]
        mov     edi, [edi]

        pushf
        push    cs
        push    offset G(callback_ret)

        push    [eax + ARG (9+7)]
        push    [eax + ARG (9+6)]
        lds     esi, [eax + ARG (9 + 0)]
        lret

GLOB(callback_ret):
        lss     esp, [esp]
        mov     ds, cs:G(ds)
        mov     esi, [esp + ARG (9+3)]
        mov     [esi], es
        mov     esi, [esp + ARG (9+2)]
        mov     [esi], edi
        mov     es, G(es)
        mov     fs, G(fs)
        mov     gs, G(gs)
        popa
        ret

        .long   G(vbe_pm_interface_end) - G(vbe_pm_interface_tab)
GLOB(vbe_pm_interface_tab):
        .word   G(vbe_pm_interface_set_window) - G(vbe_pm_interface_tab)
        .word   G(vbe_pm_interface_set_display_start) - G(vbe_pm_interface_tab)
        .word   -1              # set palette
        .word   0               # ports and memory locations
GLOB(vbe_pm_interface_set_window):
        mov     ax, 0x4f05
        int     0x10
        ret
GLOB(vbe_pm_interface_set_display_start):
        mov     ax, 0x4f07
        int     0x10
        ret
GLOB(vbe_pm_interface_end):

        .data
        .globl  G(stack_top)
        .globl  G(stack_chunk_size)

GLOB(stack_top):
        .long   0
GLOB(stack_chunk_size):
        .long   0x1000

        .data
        .globl  G(cs)
        .globl  G(ds)
        .globl  G(es)
        .globl  G(ss)
        .globl  G(fs)
        .globl  G(gs)
        .globl  G(int_trampoline)
        .globl  G(int_trampoline_nint)
        .globl  G(int_trampoline_addr)
        .globl  G(int_trampoline_end)

GLOB(int_trampoline):
        cld
        pusha
        push    ds
        push    es
        push    fs
        push    gs

        mov     eax, 0xbeef
G(ds)=.-4
        mov     ds, eax
        mov     es, G(es)
        mov     fs, G(fs)
        mov     gs, G(gs)

        xor     esi, esi
        xor     ebp, ebp
        mov     eax, ss
        mov     ecx, esp
        mov     ebx, 0xbeef
G(ss)=.-4
        cmp     eax, ebx
        je      L(dont_switch_stacks)

        lar     esi, eax
        test    esi, 0x400000
        jnz     0f

        movzx   ecx, cx
0:
        mov     esi, [G(stack_chunk_size)]
        sub     ebp, esi
        xadd    [G(stack_top)], ebp

        mov     ss, ebx
        mov     esp, ebp

L(dont_switch_stacks):
        push    eax
        push    ecx

        push    0xdeadbeef
G(int_trampoline_nint)=.-4

        push    0xdeadbeef
G(int_trampoline_addr)=.-4

        push    offset 0f
        push    offset SIG_call_int_trampoline
        ret
0:
        add     esp, 8
        lss     esp, [esp]
        add     G(stack_top), esi

        pop     gs
        pop     fs
        pop     es
        pop     ds
        popa
        iret
G(fs):   .long 0
G(gs):   .long 0
G(cs):   .long 0
G(es):   .long 0
G(int_trampoline_end)=.

# Local Variables:
# comment-start: "#"
# End:
