#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <inc/int.h>
#include <inc/mem.h>
#include <inc/dmm.h>
#include <inc/sig.h>
#include <inc/log.h>
#include <inc/mmi.h>

int key_debug_level;
FILE *dbg_fd;

static void regs_to_state (STATE__ *state, intregs r)
{
  state->eax = EAX (r);
  state->ebx = EBX (r);
  state->ecx = ECX (r);
  state->edx = EDX (r);
  state->esi = ESI (r);
  state->edi = EDI (r);
  state->ebp = EBP (r);
  state->esp = ESP (r);

  state->cs = CS (r);
  state->ds = DS (r);
  state->es = ES (r);
  /* state->ss = SS (r); */
  state->fs = FS (r);
  state->gs = GS (r);
}

static void state_to_regs (intregs r, STATE__ *state)
{
  EAX (r) = state->eax;
  EBX (r) = state->ebx;
  ECX (r) = state->ecx;
  EDX (r) = state->edx;
  ESI (r) = state->esi;
  EDI (r) = state->edi;
  EBP (r) = state->ebp;
  ESP (r) = state->esp;

  CS (r) = state->cs;
  DS (r) = state->ds;
  ES (r) = state->es;
  /* SS (r) = state->ss; */
  FS (r) = state->fs;
  GS (r) = state->gs;
}

void call_func (mm_func func, intregs r)
{
  STATE__ state;

  if (AH (r) > 0x50)
    return;

  regs_to_state (&state, r);
  if (FALSE == func (&state)) {
    Fail ("EMS call returned false");
  }
  state_to_regs (r, &state);
}

void destroy_memory_object (mach_port_t object)
{
  int was_blocked;

  was_blocked = SIG_block ();
/*   if (object->addr_size) */
/*     MEM_map_address (object->addr, object->addr, object->addr_size); */
  DMM_free (object->phys, object->size);
  xfree (object);
  SIG_unblock (was_blocked);
}

mach_port_t new_memory_object (int size)
{
  int was_blocked;
  void *phys;
  struct mach_port_s *object;

  was_blocked = SIG_block ();

  phys = DMM_alloc (size);
  if (NULL == phys)
    Fail ("DMM_alloc (%d) failed", size);

  object = xmalloc (sizeof (*object));
  memset (object, 0, sizeof (*object));
  object->phys = phys;
  object->size = size;
  SIG_unblock (was_blocked);
  return object;
}

int vm_deallocate (int unused, int address, int size)
{
  DEREF (unused);
  MEM_map_address (address, address, size);
  return KERN_SUCCESS;
}

void memory_object_remap (mach_port_t object, int offset, int addr, int size)
{
  int target;

  target = offset + (int) object->phys;
  MEM_map_address (target, addr, size);
  object->addr_size = size;
  object->addr = addr;
}

void exit_dos ()
{
  abort ();
}

void mach_error (char *msg, int code)
{
  Fail ("%s (%d)", msg, code);
}
