/*
 * This is NOT real mouse driver.
 * But it allows me to play Worms.
 * And select options in countdown.
 * So im happy.
 *
 * Score: -1
 * Bugs: sticky buttons(?)
 */

#include <stdlib.h>
#include <SDL/SDL.h>

#include <inc/asm.h>
#include <inc/sel.h>
#include <inc/mou.h>
#include <inc/eio.h>
#include <inc/pic.h>
#include <inc/int.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "MOU"
static int log_level = -1;

enum MOUSE_CONDITION {
  MOUSE_CONDITION_MOVES     = 1,
  MOUSE_CONDITION_LPRESSED  = 2,
  MOUSE_CONDITION_LRELEASED = 4,
  MOUSE_CONDITION_RPRESSED  = 8,
  MOUSE_CONDITION_RRELEASED = 16
};

static int call_mask;
static int call_seg;
static ui call_off;

static int mouse_ncom = 1;
static int mouse_nirq;
static int mouse_nport;
static int mouse_slice;

static ui data_index;
static int irqs_to_generate;

static struct mou {
  int x;
  int y;

  int x_mickey_scale;
  int y_mickey_scale;

  int x_mickey;
  int y_mickey;

  int old_x_mickey;
  int old_y_mickey;

  int lpressed;
  int rpressed;
  int lreleased;
  int rreleased;
  int moved;

  int hor_min;
  int ver_min;
  int hor_max;
  int ver_max;

  int threshold;
  ub data[3];
} mou;

static void reset ()
{
  mou.moved = 0;
  mou.lpressed = 0;
  mou.lreleased = 0;
  mou.rpressed = 0;
  mou.rreleased = 0;
}

static void encode_increments (ub xinc, ub yinc)
{
  int h2bx, h2by;
  int l6bx, l6by;

  h2bx = (xinc >> 6) & 3;
  h2by = (yinc >> 6) & 3;
  l6bx = xinc & 64;
  l6by = yinc & 64;
  mou.data[0] = (mou.data[0] & ~15) | (h2bx << 2) | (h2by);
  mou.data[1] = l6bx;
  mou.data[2] = l6by;
}

static ub sint_to_byte (Sint16 rel)
{
  ub res;

  if (rel > 127) {
    res = 0x7f;
  }
  else if (rel < -128) {
    res = 0xff;
  }
  else {
    if (rel < 0) {
      res = 0x80 | -rel;
    }
    else {
      res = rel;
    }
  }

  return res;
}

void MOU_set_callback (int mask, int seg, ui off)
{
  call_seg = seg;
  call_off = off;
  call_mask = mask;
}

void MOU_get_callback (int *mask, int *seg, ui *off)
{
  *mask = call_mask;
  *seg = call_seg;
  *off = call_off;
}

#define CLIP(v, min, max) do {                  \
  if (v > max) v = max;                         \
  else if (v < min) v = min;                    \
} while (0)

void MOU_motion_event (SDL_MouseMotionEvent *event)
{
  ub xinc, yinc;
  int xrel_mickeys, yrel_mickeys;

  ldebug ("mouse %dx%d +%d+%d\n",
          event->x, event->y,
          event->xrel, event->yrel);

  xrel_mickeys = event->xrel * mou.x_mickey_scale;
  yrel_mickeys = event->yrel * mou.y_mickey_scale;

  mou.x_mickey += xrel_mickeys;
  mou.y_mickey += yrel_mickeys;

  mou.x += xrel_mickeys;
  mou.y += yrel_mickeys;

  CLIP (mou.x, mou.hor_min, mou.hor_max);
  CLIP (mou.y, mou.ver_min, mou.ver_max);

  xinc = sint_to_byte (event->xrel);
  yinc = sint_to_byte (event->yrel);

  if (xinc || yinc)
    mou.moved = 1;
  else
    mou.moved = 0;

  encode_increments (xinc, yinc);
  irqs_to_generate = 2;
  PIC_request_service (mouse_nirq);
}

void MOU_button_event (SDL_MouseButtonEvent *event)
{
  int state_changed;
  int old_lb, old_rb;

  state_changed = 0;
  old_rb = (mou.data[0] >> 4) & 1;
  old_lb = (mou.data[0] >> 5) & 1;

  ldebug ("mouse %d, %d %dx%d\n",
          event->button, event->state,
          event->x, event->y);

  if (old_rb) {
    if ((event->state == SDL_RELEASED)
        && (event->button == SDL_BUTTON_RIGHT)) {

      mou.data[0] &= ~(1 << 4);
      state_changed = 1;

      mou.rreleased = 1;
    }
  }
  else {
    if ((event->state == SDL_PRESSED)
        && (event->button == SDL_BUTTON_RIGHT)) {

      mou.data[0] |= 1 << 4;
      state_changed = 1;

      mou.rpressed = 1;
    }
  }

  if (old_lb) {
    if ((event->state == SDL_RELEASED)
        && (event->button == SDL_BUTTON_LEFT)) {

      mou.data[0] &= ~(1 << 5);
      state_changed = 1;

      mou.lreleased = 1;
    }
  }
  else {
    if ((event->state == SDL_PRESSED)
        && (event->button == SDL_BUTTON_LEFT)) {

      mou.data[0] |= 1 << 5;
      state_changed = 1;

      mou.lpressed = 1;
    }
  }

  if (state_changed) {
    irqs_to_generate = 2;
    PIC_request_service (mouse_nirq);
  }
}

static ui read_mouse (int nport, int size)
{
  ub retval;

  DEREF (nport);
  DEREF (size);

  if (data_index > 2) {
    Aldri ();
  }

  retval = mou.data[data_index];

  if (data_index == 2)
    data_index = 0;
  else
    data_index++;

  return retval;
}

static void IRQM (INT_ARGS)
{
  DEREF (r);

  read_mouse (mouse_nport, 1);
  PIC_eoi (0);
}

static void pic_handler (int which)
{
  if (which) {
    if (irqs_to_generate) {
      PIC_request_service (mouse_nirq);
      irqs_to_generate--;
    }
    else {
      int mask;

      mask = (mou.moved  ? MOUSE_CONDITION_MOVES : 0)
        | (mou.lpressed  ? MOUSE_CONDITION_LPRESSED : 0)
        | (mou.lreleased ? MOUSE_CONDITION_LRELEASED : 0)
        | (mou.rpressed  ? MOUSE_CONDITION_RPRESSED : 0)
        | (mou.rreleased ? MOUSE_CONDITION_RRELEASED : 0);

      if (mask & call_mask) {

        int button_state;

        button_state = (mou.lpressed ? 1 : 0) | (mou.rpressed ? 2 : 0);

        reset ();
        if (0 == (call_seg | call_off))
          lwarn ("mask is set but callback function is not\n");
        else {
          struct SIG_mouse_info info;

          info.seg = call_seg;
          info.off = call_off;
          info.mask = mask;
          info.state = button_state;
          info.xpos = mou.x;
          info.ypos = mou.y;
          info.hor_mickeys = mou.x_mickey;
          info.ver_mickeys = mou.y_mickey;

          SIG_mouse_callback (&info);
        }
      }
      else reset ();
    }
  }
}

int MOU_get_state (int *x, int *y)
{
  int lb, rb;
  *x = mou.x;
  *y = mou.y;

  lb = (mou.data[0] >> 5) & 1;
  rb = (mou.data[0] >> 4) & 1;
  return lb + (rb << 1);
}

void MOU_set_position (int x, int y)
{
  mou.x = x;
  mou.y = y;
}

void MOU_hide_cursor ()
{
  int was_blocked;

  was_blocked = SIG_block ();
  SDL_ShowCursor (SDL_DISABLE);
  SIG_unblock (was_blocked);
}

void MOU_show_cursor ()
{
  int was_blocked;

  was_blocked = SIG_block ();
  SDL_ShowCursor (SDL_ENABLE);
  SIG_unblock (was_blocked);
}

void MOU_read_motion_counters (int *dx, int *dy)
{
  *dx = mou.x_mickey - mou.old_x_mickey;
  *dy = mou.y_mickey - mou.old_y_mickey;
  mou.old_x_mickey = mou.x_mickey;
  mou.old_y_mickey = mou.y_mickey;
}

void MOU_set_horizontal_range (int min, int max)
{
  mou.hor_min = min;
  mou.hor_max = max;
}

void MOU_set_vertical_range (int min, int max)
{
  mou.ver_min = min;
  mou.ver_max = max;
}

void MOU_set_mickey_ratio (int x_mickey_scale, int y_mickey_scale)
{                               /* XXX */
#if 1
  DEREF (x_mickey_scale);
  DEREF (y_mickey_scale);
#else
  mou.x_mickey_scale = x_mickey_scale;
  mou.y_mickey_scale = y_mickey_scale;
#endif
}

void MOU_set_threshold (int threshold)
{
  mou.threshold = threshold;
}

void MOU_pre_init ()
{
  static struct MUP_key keys[] = {
    {"port"  , TAG_INT, {.ip = &mouse_nport}},
    {"slice" , TAG_INT, {.ip = &mouse_slice}},
  };

  static struct MUP_group group =
    {"mouse.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void MOU_init ()
{
  log_level = MUP_get_log_level ("mou");

  switch (mouse_ncom) {
    case 1:
      mouse_nirq = 4;
      mouse_nport = 0x3f8;
      break;

    case 2:
      mouse_nirq = 3;
      mouse_nport = 0x3f8;
      break;

    default:
      Fail ("invalid mouse com port %d, must be 1 or 2", mouse_ncom);
      break;
  }

  mou.data[0] = 1 << 6;

  mou.x_mickey_scale = 2;
  mou.y_mickey_scale = 2;

  EIO_register_read_handler (mouse_nport, EIO_MASK_B, read_mouse);

  INT_set (mouse_nirq + 8, IRQM);
  PIC_register (mouse_nirq, pic_handler);
}
