#include <inc/mem.h>
#include <inc/asm.h>
#include <inc/sel.h>
#include <inc/ldt.h>
#include <inc/log.h>
#include <inc/mup.h>

static int user_code, user_data;

us SEL_get_user_code ()
{
  return user_code;
}

us SEL_get_user_data ()
{
  return user_data;
}

ui SEL_get_selector_limit (int selector)
{
  if (SEL_is_system_selector (selector))
    return ~0U;
  else
    return LDT_get_selector_limit (selector);
}

unsigned long SEL_get_selector_base (int selector)
{
  if (SEL_is_system_selector (selector))
    return 0;
  else
    return LDT_get_selector_base (selector);
}

unsigned long SEL_linear_addr (int selector, ui off)
{
  if (SEL_is_system_selector (selector))
    return off;
  else {
    unsigned long base;

    base = LDT_get_selector_base (selector);
    if (LDT_is_selector_big (selector))
      return base + off;
    else
      return base + (off & 0xffff);
  }
}

int SEL_is_system_selector (int selector)
{
  return (user_code == selector) || (user_data == selector);
}

void SEL_init ()
{
  user_code = LDT_create_system_alias (1, 0);
  if (-1 == user_code) {
    Fail ("failed to allocate user code selector");
  }

  user_data = LDT_create_system_alias (0, 0);
  if (-1 == user_data) {
    Fail ("failed to allocate user data selector");
  }

  LDT_make_selector_immutable (user_code);
  LDT_make_selector_immutable (user_data);
}
