#include <time.h>

#include <inc/tid.h>
#include <inc/mup.h>
#include <inc/log.h>

#define LOG_PREFIX "TID"
static int log_level;

static ul base, bias, paused_at;
static time_t base_sec;

static inline ul now ()
{
  ul now;
  struct timespec ts;

  if (-1 == clock_gettime (CLOCK_PROCESS_CPUTIME_ID, &ts)) {
    ERRFail ("clock_gettime failed");
  }

  now = ts.tv_sec * (ul) TID_NSEC + ts.tv_nsec;
  return now + bias;
}

ul TID_get_abs_nsec ()
{
  return now ();
}

ul TID_get_nsec_diff (ul ref)
{
  return now () - ref;
}

void TID_get_tm (struct TID_tm *tm)
{
  ul nsec;
  time_t t;
  struct tm ttm;
  struct tm *rtm;

  nsec = now () - base;
  t = (time_t) (base_sec + (nsec / (ul) TID_NSEC));
  rtm = localtime_r (&t, &ttm);

  if (NULL == rtm) {
    ERRFail ("localtime_r failed");
  }

  tm->nsec = (nsec + base_sec * (ul) TID_NSEC) - (t * (ul) TID_NSEC);
  tm->sec = rtm->tm_sec;
  tm->min = rtm->tm_min;
  tm->hour = rtm->tm_hour;
  tm->mday = rtm->tm_mday;
  tm->wday = rtm->tm_wday;
  tm->mon = rtm->tm_mon + 1;
  tm->year = rtm->tm_year + 1900;
}

void TID_pause ()
{
  paused_at = now ();
}

void TID_resume ()
{
  bias = now () - paused_at;
}

void TID_init ()
{
  DEREF (base);
  DEREF (log_level);

  log_level = MUP_get_log_level ("tid");
  base = now ();
  base_sec = time (NULL);
}
