set verbose off
set confirm off
set history save on
set print symbol-filename off
set print max-symbolic-offset 2
set print pretty on
set step-mode on
set pagination off

define regs
printf "eax = %08x ebx = %08x ecx = %08x edx = %08x\n", $eax, $ebx, $ecx, $edx
printf "esi = %08x edi = %08x ebp = %08x esp = %08x\n", $esi, $edi, $ebp, $esp
printf "cs = %04x ds = %04x es = %04x fs = %04x gs = %04x ss = %04x\n", \
$cs, $ds, $es, $fs, $gs, $ss
printf "eip = %08x elfags = %08x\n", $pc, $eflags
end

define xseg
  if (0 == ($arg1 & 4))
     set $_32bit=1
     set $_base=0
  else
     set $_32bit=LDT_tab[$arg1>>3].seg_32bit
     set $_base=LDT_tab[$arg1>>3].base_addr
  end

  set $_addr = $_base + $arg2
  if !$_32bit
     set architecture i8086
  end

  x/$arg0 $_addr

  if !$_32bit
     set architecture i386
  end
end

define xss
xseg $arg0 $ss $arg1
end

define xds
xseg $arg0 $ds $arg1
end

define xes
xseg $arg0 $es $arg1
end

define xcs
xseg $arg0 $cs $arg1
end

define xpc
xcs $arg0 $pc
end

define xsp
xss $arg0 $esp
end

define xrseg
set $_addr = ($arg1 * 16 + $arg2)
set architecture i8086
x/$arg0 $_addr
set architecture i386
end

define xrcs
xrseg $arg0 vm86p.regs.cs
end

define xres
xrseg $arg0 vm86p.regs.es
end

define xrss
xrseg $arg0 vm86p.regs.ss
end

define xrds
xrseg $arg0 vm86p.regs.ds
end

define xrfs
xrseg $arg0 vm86p.regs.fs
end

define xrgs
xrseg $arg0 vm86p.regs.gs
end

define xrpc
xrcs $arg0 $pc
end

define xrsp
xrseg $arg0 vm86p.regs.ss vm86p.regs.esp
end

define print_regs
       printf "eax = %08x ebx = %08x ecx = %08x edx = %08x\n", \
              $arg0.eax, $arg0.ebx, $arg0.ecx, $arg0.edx
       printf "esi = %08x edi = %08x ebp = %08x esp = %08x\n", \
	      $arg0.esi, $arg0.edi, $arg0.ebp, $arg0.esp
       printf "cs = %04x ds = %04x es = %04x fs = %04x gs = %04x ss = %04x\n", \
	      $arg0.cs, $arg0.ds, $arg0.es, $arg0.fs, $arg0.gs, $arg0.ss
       printf "eip = %08x elfags = %08x\n", $arg0.eip, $arg0.eflags
end

define print_regs
       printf "eax = %08x ebx = %08x ecx = %08x edx = %08x\n", \
              $arg0.eax, $arg0.ebx, $arg0.ecx, $arg0.edx
       printf "esi = %08x edi = %08x ebp = %08x esp = %08x\n", \
	      $arg0.esi, $arg0.edi, $arg0.ebp, $arg0.esp
       printf "cs = %04x ds = %04x es = %04x fs = %04x gs = %04x ss = %04x\n", \
	      $arg0.cs, $arg0.ds, $arg0.es, $arg0.fs, $arg0.gs, $arg0.ss
       printf "eip = %08x elfags = %08x\n", $arg0.eip, $arg0.eflags
end

define vregs
print_regs vm86p.regs
end

define cregs
print_regs context
end

define trampret
  set $_ret_ss = *(us *)((ui)$esp+12)
  set $_ret_sp = *(ui *)((ui)$esp+8)
  if (0 == ($_ret_ss & 4))
     set $_32bit=1
     set $_base=0
  else
     set $_32bit=LDT_tab[$_ret_ss>>3].seg_32bit
     set $_base=LDT_tab[$_ret_ss>>3].base_addr
  end
  set $_ret_cs = *(us *)($_base+$_ret_sp+12*4+4)
  set $_ret_ip = *(ui *)($_base+$_ret_sp+12*4)
  p $_ret_cs
  p $_ret_ip
end

define dr
xrseg 20i vm86p.regs.cs vm86p.regs.eip
end

define dc
xseg 20i context.cs context.eip
end

define drs
if in.vm86
   dr
else
   if (in.x86)
      set $base=(ui)current->env.segs[_SEG_CS].base
      set $seg32=current->env.segs[_SEG_CS].flags&0x400000
      set $xip=(ui)current->env.eip
   else
      if ($cs & 4) == 4
           set $base=LDT_tab[$cs>>3].base_addr
	   set $seg32=LDT_tab[$cs>>3].seg_32bit
      else
	   set $base=0
	   set $seg32=1
      end
   end
   set $xip=$eip
   if $seg32
      x/20i $base+$xip
   else
      xrseg 20i 0 ($base+$xip)
   end
end
end

define sdd
stepi
regs
xcs 20i $eip
end

define dump
printf "eax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n", \
       $arg0->regs[_REG_EAX].d, $arg0->regs[_REG_EBX].d, \
       $arg0->regs[_REG_ECX].d, $arg0->regs[_REG_EDX].d

printf "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n", \
       $arg0->regs[_REG_ESI].d, $arg0->regs[_REG_EDI].d, \
       $arg0->regs[_REG_EBP].d, $arg0->regs[_REG_ESP].d

printf "eip=%#010x efl=%#010x\n", $arg0->eip, $arg0->eflags
printf "cs=%#06x ds=%#06x es=%#06x ss=%#06x fs=%#06x gs=%#06x\n\n", \
       $arg0->segs[_SEG_CS].selector, $arg0->segs[_SEG_DS].selector, \
       $arg0->segs[_SEG_ES].selector, $arg0->segs[_SEG_SS].selector, \
       $arg0->segs[_SEG_FS].selector, $arg0->segs[_SEG_GS].selector
end

define dumpvm
printf "eax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n", \
       $arg0->eax, $arg0->ebx, $arg0->ecx, $arg0->edx
printf "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n",
       $arg0->esi, $arg0->edi, $arg0->ebp, $arg0->esp
printf "eip=%#010x efl=%#010x\n",  $arg0->eip, $arg0->eflags
printf "cs=%#06x ds=%#06x es=%#06x ss=%#06x fs=%#06x gs=%#06x\n\n", \
        $arg0->cs, $arg0->ds, $arg0->es, $arg0->ss, $arg0->fs, $arg0->gs
end

define dregs
dump current->env
end
